// Copyright (C) 2013 Jann Horn <jann@thejh.net>
// This file is licensed under the GNU GPL v2 (see
// the LICENSE file).

#undef __KERNEL__
#define __KERNEL__
#undef MODULE
#define MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/uaccess.h>
#include <linux/miscdevice.h>

struct fetch_fops_args {
  int fd;
  u64 retp;
};

#define MOCTEL_FETCH_FOPS _IOR('m', 1, struct fetch_fops_args)

static int ioctl_open(struct inode *nodp, struct file *filp) {
  return 0;
}

static const struct file_operations dummy_fops = {};

static long ioctl_ioctl(struct file *filp_, unsigned int cmd, unsigned long arg) {
  struct fetch_fops_args args;
  struct file *filp;
  const struct file_operations *fops;

  if (cmd != MOCTEL_FETCH_FOPS) return -EINVAL;
  if (copy_from_user(&args, (struct fetch_fops_args __user *)arg, sizeof(args))) return -EINVAL;
  filp = fget(args.fd);
  if (!filp) return -EBADF;
  fops = filp->f_op;
  if (!fops) fops = &dummy_fops;
  fput(filp);
  if (copy_to_user((struct file_operations __user *)args.retp, fops, sizeof(*fops))) return -EINVAL;
  return 0;
}

static int ioctl_release(struct inode *nodp, struct file *filp) {
  return 0;
}

static const struct file_operations ioctl_fops = {
  .owner = THIS_MODULE,
  .unlocked_ioctl = ioctl_ioctl,
  .open = ioctl_open,
  .release = ioctl_release
};

static struct miscdevice ioctl_miscdev = {
  .minor = 0,
  .name = "ioctl_info",
  .fops = &ioctl_fops
};

static int __init init_ioctl(void) {
  int ret;
  printk(KERN_INFO "loading ioctl helper\n");

  ret = misc_register(&ioctl_miscdev);
  return ret;
}

static void __exit cleanup_ioctl(void) {
  printk(KERN_INFO "unloading ioctl helper\n");
  misc_deregister(&ioctl_miscdev);
}

module_init(init_ioctl);
module_exit(cleanup_ioctl);
